/*
  XLiFE++ is an extended library of finite elements written in C++
  Copyright (C) 2014  Lunéville, Eric; Kielbasiewicz, Nicolas; Lafranche, Yvon; Nguyen, Manh-Ha; Chambeyron, Colin

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
  \file geometries2D.hpp
  \authors Y. Lafranche, N. Kielbasiewicz
  \since 18 oct 2012
  \date 30 jul 2015

  \brief Definition of the classes corresponding to 2D canonical geometrical shapes.

  It is quite easy to find in which category a geometry is, according to their mathematical
  definition, but they are some exceptions:
  - Cylinder and Cone are surfaces but can also represent volumes inside
  - Ellipse is a closed curve or the surface inside (whereas there is the
    distinction circle / disk)
  - Ellipsoid is a closed surface or the volume inside (whereas there is the
    distinction Sphere / Ball)

  It becomes all the more tricky than if we respect the geometrical hierarchy, we will derive a
  prism (semantically 3D) from a cylinder (2D or 3D) or a pyramid (semantically 3D) from a cone.
  Another example is the general definition of a cylinder, from a surface and an axis: it supposes
  that the basis is a surface. But ellipses are suposed to be curves !!!

  This is the reason why we chose to defines ellipses as surfaces, ellipsoids as volumes (coherence) and cylinders
  and cones as volumes.

*/

#ifndef GEOMETRIES_2D_HPP
#define GEOMETRIES_2D_HPP

#include "Geometry.hpp"

namespace xlifepp
{
//============================================================================================
/*!
  \class Surface
  base class for 2D geometries
*/
//============================================================================================
class Surface : public Geometry
{
  public:
    //! default constructor for 2D geometries
    Surface();

  protected:

    std::vector<Point> p_;    //!< construction points including true vertices
    std::vector<real_t> h_;   //!< local mesh step on each vertex of the surface
    std::vector<number_t> n_; //!< number of nodes on each edge

    //@{
    //! true constructor functions
    void buildParam(const Parameter& p);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    virtual void setParametrization()
    {error("free_error","no setParametrization defined");}

    //@}

  public:
    Surface(const Surface& s) : Geometry(s), p_(s.p_), h_(s.h_), n_(s.n_) {} //!< copy constructor
    virtual Geometry* clone() const { return new Surface(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Surface(*this); } //!< virtual copy constructor for Surface
    virtual ~Surface() {} //!< virtual destructor

    //@{
     //! accessor to number of nodes per edge and discretized steps
    const std::vector<number_t>& n() const {return n_;}
    std::vector<number_t>& n() {return n_;}
    number_t n(number_t i) const { return n_[i-1];}
    number_t& n(number_t i) { return n_[i-1];}
    const std::vector<real_t>& h() const {return h_;}
    std::vector<real_t>& h() {return h_;}
    real_t h(number_t i) const { return h_[i-1];}
    real_t& h(number_t i) { return h_[i-1];}
    //@}

    //@{
    //! accessor to nodes/vertices
    const std::vector<Point>& p() const { return p_; }
    std::vector<Point>& p() { return p_; }
    const Point& p(number_t i) const { return p_[i-1]; }
    Point& p(number_t i) { return p_[i-1]; }
    void setHstep(real_t hs) {h_=std::vector<real_t>(p_.size(),hs);}
    bool withNnodes() const { return (h_.size() == 0); }
    virtual void setNnodes(number_t i) {number_t d=n_.size();n_.clear();n_.resize(d,i);}
    //@}

    const Surface* surface() const //! access to Surface object (const), virtual in Geometry
    { return this; }
    Surface* surface()  //! access to Surface object (non const), virtual in Geometry
    { return this; }

    //! test if a surface is a polygon
    bool isPolygon() const
    {return (shape_ == _polygon || shape_ == _triangle || shape_ == _quadrangle || shape_ == _parallelogram || shape_ == _rectangle || shape_ == _square);}

    virtual bool isTranslated(const Geometry&, Point&) const;//! check if points of current surface are points of a translated points of an other surface
    virtual void computeBB();                        //! computes the bounding box (in x,y,z)
    virtual void computeMB()                         //! computes the minimal box (default=bounding box)
    { minimalBox=MinimalBox(boundingBox.bounds());}

    //! format as string
    virtual string_t asString() const { error("shape_not_handled", words("shape",shape_)); return string_t(); }

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a Surface
    virtual Surface& transform(const Transformation& t);
    //! apply a translation on a Surface (vector version)
    virtual Surface& translate(std::vector<real_t> u = std::vector<real_t>(3,0.))
      {return transform(Translation(u));}
    //! apply a translation on a Surface (3 reals version)
    virtual Surface& translate(real_t ux, real_t uy = 0., real_t uz = 0.)
      {return transform(Translation(Point(ux,uy,uz)));}
    //! apply a rotation on a Surface
    virtual Surface& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.)
      {return transform(Rotation2d(c,angle));}
    //! apply a rotation on a Surface
    virtual Surface& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.)
      {return transform(Rotation3d(c,d,angle));}
    //! apply a rotation on a Surface
    virtual Surface& rotate3d(real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,0.),angle));}
    //! apply a rotation on a Surface
    virtual Surface& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,dz),angle));}
    //! apply a rotation on a Surface
    virtual Surface& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,0.),angle));}
    //! apply a rotation on a Surface
    virtual Surface& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,dz),angle));}
    //! apply a homothety on a Surface
    virtual Surface& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.)
      {return transform(Homothety(c,factor));}
    //! apply a homothety on a Surface
    virtual Surface& homothetize(real_t factor)
     {return transform(Homothety(Point(0.,0.,0.),factor));}
    //! apply a point reflection on a Surface
    virtual Surface& pointReflect(const Point& c = Point(0.,0.,0.))
       {return transform(PointReflection(c));}
    //! apply a reflection2d on a Surface
    virtual Surface& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.))
       {return transform(Reflection2d(c,d));}
    //! apply a reflection2d on a Surface
    virtual Surface& reflect2d(const Point& c, real_t dx, real_t dy = 0.)
      {return transform(Reflection2d(c,Point(dx,dy)));}
    //! apply a reflection3d on a Surface
    virtual Surface& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.))
       {return transform(Reflection3d(c,n));}
    //! apply a reflection3d on a Surface
    virtual Surface& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.)
     {return transform(Reflection3d(c,Point(nx,ny,nz)));}
};

//============================================================================================
/*!
  \class Polygon
  definition of a polygonal geometry in R^3

  Polygon constructors are based on a key-value system. Here are the available keys:
  - _vertices: the vertices defining the polygon, oriented along the boundary (clockwise or counterclockwise)
  - _nnodes: to define the number of nodes on each edge of the Polygon
  - _hsteps: to define the local mesh steps on the vertices of the Polygon
  - _domain_name: to define the domain name
  - _side_names: to define the side names
  - _varnames: to define the variable names for print purpose
 */
//============================================================================================
class Polygon : public Surface
{
  public:
    //! default constructor
    Polygon();
    //! default constructor with vertices
    Polygon(const std::vector<Point>& vertices, const std::vector<number_t>& n=std::vector<number_t>(4,2),
            const string_t& domName=string_t());
    //! default constructor with vertices
    Polygon(const std::vector<Point>& vertices, const std::vector<real_t>& h, const string_t& domName=string_t());
    //! constructor with 1 Parameter
    Polygon(Parameter p1);
    //! constructor with 2 Parameter
    Polygon(Parameter p1, Parameter p2);
    //! constructor with 3 Parameter
    Polygon(Parameter p1, Parameter p2, Parameter p3);
    //! constructor with 4 Parameter
    Polygon(Parameter p1, Parameter p2, Parameter p3, Parameter p4);

  protected:
    //@{
    //! true constructor functions
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    virtual void setParametrization();
    //@}

  public:
    //! destructor
    virtual ~Polygon() {}

    //! format as string
    virtual string_t asString() const;
    void printDetail(std::ostream& os) const; //!< print more infos

    virtual Geometry* clone() const { return new Polygon(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Polygon(*this); } //!< virtual copy constructor for Surface
    virtual Polygon* clonePG() const { return new Polygon(*this); } //!< virtual copy constructor for Polygon

    std::vector<int_t> nnodesPerBorder(); //!< i-th value is nnodes of the Geometry's i-th border, else -1 if defined with hsteps

    virtual std::vector<const Point*> boundNodes() const; //!< returns list of nodes on boundary (const)
    virtual std::vector<Point*> nodes(); //!< returns list of every node (non const)
    virtual std::vector<const Point*> nodes() const;    //!< returns list of every node (const)
    virtual std::vector<const Point*> vertices() const; //!< returns list of vertices (const)
    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > curves() const; //!< return list of edges (const)
    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > surfs() const; //!< return list of surfaces (const)
    virtual number_t nbSides() const { return p_.size(); } //!< returns the number of sides
    virtual void collect(const string_t& n, std::list<Geometry*>&) const; //!< collect in a list all canonical geometry's with name n
    virtual Point centroid() const; //!< return centroid of polygon
    virtual bool isPlane() const {return true;}

    //! computes the minimal box
    virtual void computeMB();
    //! access to child Polygon object (const)
    virtual const Polygon* polygon() const {return this;}
    //! access to child Polygon object
    virtual Polygon* polygon() {return this;}
    //! create boundary geometry
    virtual Geometry& buildBoundary() const;

    //------------------------------------------------
    //         transformations facilities
    //------------------------------------------------
    //! apply a geometrical transformation on a Polygon
    virtual Polygon& transform(const Transformation& t);
    //! apply a translation on a Polygon (vector version)
    virtual Polygon& translate(std::vector<real_t> u = std::vector<real_t>(3,0.))
      {return transform(Translation(u));}
    //! apply a translation on a Polygon (3 reals version)
    virtual Polygon& translate(real_t ux, real_t uy = 0., real_t uz = 0.)
      {return transform(Translation(Point(ux,uy,uz)));}
    //! apply a rotation on a Polygon
    virtual Polygon& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.)
      {return transform(Rotation2d(c,angle));}
    //! apply a rotation on a Polygon
    virtual Polygon& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.)
      {return transform(Rotation3d(c,d,angle));}
    //! apply a rotation on a Polygon
    virtual Polygon& rotate3d(real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,0.),angle));}
    //! apply a rotation on a Polygon
    virtual Polygon& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,dz),angle));}
    //! apply a rotation on a Polygon
    virtual Polygon& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,0.),angle));}
    //! apply a rotation on a Polygon
    virtual Polygon& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,dz),angle));}
    //! apply a homothety on a Polygon
    virtual Polygon& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.)
      {return transform(Homothety(c,factor));}
    //! apply a homothety on a Polygon
    virtual Polygon& homothetize(real_t factor)
     {return transform(Homothety(Point(0.,0.,0.),factor));}
    //! apply a point reflection on a Polygon
    virtual Polygon& pointReflect(const Point& c = Point(0.,0.,0.))
       {return transform(PointReflection(c));}
    //! apply a reflection2d on a Polygon
    virtual Polygon& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.))
       {return transform(Reflection2d(c,d));}
    //! apply a reflection2d on a Polygon
    virtual Polygon& reflect2d(const Point& c, real_t dx, real_t dy = 0.)
      {return transform(Reflection2d(c,Point(dx,dy)));}
    //! apply a reflection3d on a Polygon
    virtual Polygon& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.))
       {return transform(Reflection3d(c,n));}
    //! apply a reflection3d on a Polygon
    virtual Polygon& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.)
     {return transform(Reflection3d(c,Point(nx,ny,nz)));}
};

//============================================================================================
/*!
  \class Triangle
  definition of a triangular geometry in R^3

  Triangle constructors are based on a key-value system. Here are the available keys:
  - _v1, _v2, _v3: to define the vertices of the triangle, counterclockwise oriented
  - _nnodes: to define the number of nodes on each edge of the Triangle
  - _hsteps: to define the local mesh steps on the vertices of the Triangle
  - _domain_name: to define the domain name
  - _side_names: to define the side names
  - _varnames: to define the variable names for print purpose
 */
//============================================================================================
class Triangle : public Polygon
{
  public:
    //! default constructor
    Triangle();
    //! default constructor with 3 Point
    Triangle(const Point& p1, const Point& p2, const Point& p3, const std::vector<number_t>& n=std::vector<number_t>(3,2),
             const string_t& domName=string_t());
    //! default constructor with 3 Point
    Triangle(const Point& p1, const Point& p2, const Point& p3, const std::vector<real_t>& h, const string_t& domName=string_t());
    //! constructor with 3 Parameter
    Triangle(Parameter p1, Parameter p2, Parameter p3);
    //! constructor with 4 Parameter
    Triangle(Parameter p1, Parameter p2, Parameter p3, Parameter p4);
    //! constructor with 5 Parameter
    Triangle(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5);
    //! constructor with 6 Parameter
    Triangle(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5,
             Parameter p6);

  private:
    //@{
    //! true constructor functions
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    void setParametrization();
    //@}

  public:
    //! destructor
    virtual ~Triangle() {}

    virtual Geometry* clone() const { return new Triangle(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Triangle(*this); } //!< virtual copy constructor for Surface
    virtual Polygon* clonePG() const { return new Triangle(*this); } //!< virtual copy constructor for Polygon

    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > surfs() const; //!< return list of surfaces (const)
    virtual number_t nbSides() const { return 3; } //!< returns the number of sides
    real_t measure() const;

    //! access to child Triangle object (const)
    virtual const Triangle* triangle() const {return this;}
    //! access to child Triangle object
    virtual Triangle* triangle() {return this;}

    //! format as string
    virtual string_t asString() const;

    //------------------------------------------------
    //         transformations facilities
    //------------------------------------------------
    //! apply a geometrical transformation on a Triangle
    virtual Triangle& transform(const Transformation& t);
    //! apply a translation on a Triangle (vector version)
    virtual Triangle& translate(std::vector<real_t> u = std::vector<real_t>(3,0.))
      {return transform(Translation(u));}
    //! apply a translation on a Triangle (3 reals version)
    virtual Triangle& translate(real_t ux, real_t uy = 0., real_t uz = 0.)
      {return transform(Translation(Point(ux,uy,uz)));}
    //! apply a rotation on a Triangle
    virtual Triangle& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.)
      {return transform(Rotation2d(c,angle));}
    //! apply a rotation on a Triangle
    virtual Triangle& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.)
      {return transform(Rotation3d(c,d,angle));}
    //! apply a rotation on a Triangle
    virtual Triangle& rotate3d(real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,0.),angle));}
    //! apply a rotation on a Triangle
    virtual Triangle& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,dz),angle));}
    //! apply a rotation on a Triangle
    virtual Triangle& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,0.),angle));}
    //! apply a rotation on a Triangle
    virtual Triangle& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,dz),angle));}
    //! apply a homothety on a Triangle
    virtual Triangle& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.)
      {return transform(Homothety(c,factor));}
    //! apply a homothety on a Triangle
    virtual Triangle& homothetize(real_t factor)
     {return transform(Homothety(Point(0.,0.,0.),factor));}
    //! apply a point reflection on a Triangle
    virtual Triangle& pointReflect(const Point& c = Point(0.,0.,0.))
       {return transform(PointReflection(c));}
    //! apply a reflection2d on a Triangle
    virtual Triangle& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.))
       {return transform(Reflection2d(c,d));}
    //! apply a reflection2d on a Triangle
    virtual Triangle& reflect2d(const Point& c, real_t dx, real_t dy = 0.)
      {return transform(Reflection2d(c,Point(dx,dy)));}
    //! apply a reflection3d on a Triangle
    virtual Triangle& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.))
       {return transform(Reflection3d(c,n));}
    //! apply a reflection3d on a Triangle
    virtual Triangle& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.)
     {return transform(Reflection3d(c,Point(nx,ny,nz)));}

    Vector<real_t> funParametrization(const Point& pt, Parameters& pars, DiffOpType d=_id) const;     //!< barycentric parametrization (1-u-v)*p1+u*p2+v*p3
    Vector<real_t> invParametrization(const Point& pt, Parameters& pars, DiffOpType d=_id) const;     //!< inverse of barycentric parametrization (u,v)=invf(p)
    Vector<real_t> duffyParametrization(const Point& pt, Parameters& pars, DiffOpType d=_id) const;   //!< duffy parametrization on [0,1]x[0,1]
    Vector<real_t> invDuffyParametrization(const Point& pt, Parameters& pars, DiffOpType d=_id) const;//!< inverse of duffy parametrization u,v)=invf(p)
};

inline Vector<real_t> parametrization_Triangle(const Point& pt, Parameters& pars, DiffOpType d=_id)    //! extern parametrization call (Duffy)
{return reinterpret_cast<const Triangle*>(pars("geometry").get_p())->duffyParametrization(pt,pars,d);}
inline Vector<real_t> invParametrization_Triangle(const Point& pt, Parameters& pars, DiffOpType d=_id) //! extern invParametrization call (Duffy)
{return reinterpret_cast<const Triangle*>(pars("geometry").get_p())->invDuffyParametrization(pt,pars,d);}

//============================================================================================
/*!
  \class Quadrangle
  definition of a quadrangular geometry in R^3

  Quadrangle constructors are based on a key-value system. Here are the available keys:
  - _v1, _v2, _v3, _v4: to define the vertices of the Quadrangle, counterclockwise oriented
  - _nnodes: to define the number of nodes on each edge of the Quadrangle
  - _hsteps: to define the local mesh steps on the vertices of the Quadrangle
  - _domain_name: to define the domain name
  - _side_names: to define the side names
  - _varnames: to define the variable names for print purpose
 */
//============================================================================================
class Quadrangle : public Polygon
{
  public:
    //! default constructor
    Quadrangle();
    //! default constructor with 4 points
    Quadrangle(const Point& p1, const Point& p2, const Point& p3, const Point& p4,
               const std::vector<number_t>& n=std::vector<number_t>(4,2), const string_t& domName=string_t());
    //! default constructor with 4 points
    Quadrangle(const Point& p1, const Point& p2, const Point& p3, const Point& p4,
               const std::vector<real_t>& h, const string_t& domName=string_t());
    //! constructor with 4 Parameter
    Quadrangle(Parameter p1, Parameter p2, Parameter p3, Parameter p4);
    //! constructor with 5 Parameter
    Quadrangle(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5);
    //! constructor with 6 Parameter
    Quadrangle(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5,
             Parameter p6);
    //! constructor with 7 Parameter
    Quadrangle(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5,
             Parameter p6, Parameter p7);

  protected:
    //@{
    //! true constructor functions
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    virtual void setParametrization();
    //@}

  public:
    //! destructor
    virtual ~Quadrangle() {}

    virtual Geometry* clone() const { return new Quadrangle(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Quadrangle(*this); } //!< virtual copy constructor for Surface
    virtual Polygon* clonePG() const { return new Quadrangle(*this); } //!< virtual copy constructor for Polygon

    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > surfs() const; //!< returns list of surfaces (const)
    virtual number_t nbSides() const { return 4; } //!< returns the number of sides
    real_t measure() const;

    //! access to child Quadrangle object (const)
    virtual const Quadrangle* quadrangle() const {return this;}
    //! access to child Quadrangle object
    virtual Quadrangle* quadrangle() {return this;}

    //! format as string
    virtual string_t asString() const;

    //------------------------------------------------
    //         transformations facilities
    //------------------------------------------------
    //! apply a geometrical transformation on a Quadrangle
    virtual Quadrangle& transform(const Transformation& t);
    //! apply a translation on a Quadrangle (vector version)
    virtual Quadrangle& translate(std::vector<real_t> u = std::vector<real_t>(3,0.))
      {return transform(Translation(u));}
    //! apply a translation on a Quadrangle (3 reals version)
    virtual Quadrangle& translate(real_t ux, real_t uy = 0., real_t uz = 0.)
      {return transform(Translation(Point(ux,uy,uz)));}
    //! apply a rotation on a Quadrangle
    virtual Quadrangle& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.)
      {return transform(Rotation2d(c,angle));}
    //! apply a rotation on a Quadrangle
    virtual Quadrangle& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.)
      {return transform(Rotation3d(c,d,angle));}
    //! apply a rotation on a Quadrangle
    virtual Quadrangle& rotate3d(real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,0.),angle));}
    //! apply a rotation on a Quadrangle
    virtual Quadrangle& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,dz),angle));}
    //! apply a rotation on a Quadrangle
    virtual Quadrangle& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,0.),angle));}
    //! apply a rotation on a Quadrangle
    virtual Quadrangle& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,dz),angle));}
    //! apply a homothety on a Quadrangle
    virtual Quadrangle& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.)
      {return transform(Homothety(c,factor));}
    //! apply a homothety on a Quadrangle
    virtual Quadrangle& homothetize(real_t factor)
     {return transform(Homothety(Point(0.,0.,0.),factor));}
    //! apply a point reflection on a Quadrangle
    virtual Quadrangle& pointReflect(const Point& c = Point(0.,0.,0.))
       {return transform(PointReflection(c));}
    //! apply a reflection2d on a Quadrangle
    virtual Quadrangle& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.))
       {return transform(Reflection2d(c,d));}
    //! apply a reflection2d on a Quadrangle
    virtual Quadrangle& reflect2d(const Point& c, real_t dx, real_t dy = 0.)
      {return transform(Reflection2d(c,Point(dx,dy)));}
    //! apply a reflection3d on a Quadrangle
    virtual Quadrangle& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.))
       {return transform(Reflection3d(c,n));}
    //! apply a reflection3d on a Quadrangle
    virtual Quadrangle& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.)
     {return transform(Reflection3d(c,Point(nx,ny,nz)));}

    Vector<real_t> funParametrization(const Point& pt, Parameters& pars, DiffOpType d=_id) const; //!< parametrization (1-u-v)*p1+u*p2+v*p4-uv(p2-p1+p4-p3)
    Vector<real_t> invParametrization(const Point& pt, Parameters& pars, DiffOpType d=_id) const; //!< inverse of parametrization (u,v)=invf(p)
};

inline Vector<real_t> parametrization_Quadrangle(const Point& pt, Parameters& pars, DiffOpType d=_id)    //! extern parametrization call
{return reinterpret_cast<const Quadrangle*>(pars("geometry").get_p())->funParametrization(pt,pars,d);}
inline Vector<real_t> invParametrization_Quadrangle(const Point& pt, Parameters& pars, DiffOpType d=_id) //! extern invParametrization call
{return reinterpret_cast<const Quadrangle*>(pars("geometry").get_p())->invParametrization(pt,pars,d);}

//============================================================================================
/*!
  \class Parallelogram
  definition of a parallelogram geometry in R^3

  Parallelogram constructors are based on a key-value system. Here are the available keys:
  - _v1, _v2, _v3, _v4: to define the vertices of the Parallelogram, counterclockwise oriented (_v3 is optional)
  - _nnodes: to define the number of nodes on each edge of the Parallelogram
  - _hsteps: to define the local mesh steps on the vertices of the Parallelogram
  - _domain_name: to define the domain name
  - _side_names: to define the side names
  - _varnames: to define the variable names for print purpose
 */
//============================================================================================
class Parallelogram : public Quadrangle
{
  public:
    //! default constructor
    Parallelogram();
    //! default constructor with 3 points
    Parallelogram(const Point& p1, const Point& p2, const Point& p4,
                  const std::vector<number_t>& n=std::vector<number_t>(4,2), const string_t& domName=string_t());
    //! default constructor with 3 points
    Parallelogram(const Point& p1, const Point& p2, const Point& p4,
                  const std::vector<real_t>& h, const string_t& domName=string_t());
    //! constructor with 3 Parameter
    Parallelogram(Parameter p1, Parameter p2, Parameter p3);
    //! constructor with 4 Parameter
    Parallelogram(Parameter p1, Parameter p2, Parameter p3, Parameter p4);
    //! constructor with 5 Parameter
    Parallelogram(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5);
    //! constructor with 6 Parameter
    Parallelogram(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5,
             Parameter p6);

  protected:
    //@{
    //! true constructor functions
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    virtual void setParametrization();
    //@}

  public:
    //! destructor
    virtual ~Parallelogram() {}

    //! returns the number of nodes on first or third edge
    number_t n1() const {return n_[0];}
    //! returns the number of nodes on second or fourth edge
    number_t n2() const {return n_[1];}

    //! format as string
    virtual string_t asString() const;

    virtual Geometry* clone() const { return new Parallelogram(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Parallelogram(*this); } //!< virtual copy constructor for Surface
    virtual Polygon* clonePG() const { return new Parallelogram(*this); } //!< virtual copy constructor for Polygon

    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > surfs() const; //!< returns list of surfaces (const)
    real_t length1() const; //!< length of first or third side
    real_t length2() const; //!< length of second or fourth side
    real_t measure() const; //!< surface of the parallelogram

    //! access to child Parallelogram object (const)
    virtual const Parallelogram* parallelogram() const {return this;}
    //! access to child Parallelogram object
    virtual Parallelogram* parallelogram() {return this;}

    //------------------------------------------------
    //         transformations facilities
    //------------------------------------------------
    //! apply a geometrical transformation on a Parallelogram
    virtual Parallelogram& transform(const Transformation& t);
    //! apply a translation on a Parallelogram (vector version)
    virtual Parallelogram& translate(std::vector<real_t> u = std::vector<real_t>(3,0.))
      {return transform(Translation(u));}
    //! apply a translation on a Parallelogram (3 reals version)
    virtual Parallelogram& translate(real_t ux, real_t uy = 0., real_t uz = 0.)
      {return transform(Translation(Point(ux,uy,uz)));}
    //! apply a rotation on a Parallelogram
    virtual Parallelogram& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.)
      {return transform(Rotation2d(c,angle));}
    //! apply a rotation on a Parallelogram
    virtual Parallelogram& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.)
      {return transform(Rotation3d(c,d,angle));}
    //! apply a rotation on a Parallelogram
    virtual Parallelogram& rotate3d(real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,0.),angle));}
    //! apply a rotation on a Parallelogram
    virtual Parallelogram& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,dz),angle));}
    //! apply a rotation on a Parallelogram
    virtual Parallelogram& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,0.),angle));}
    //! apply a rotation on a Parallelogram
    virtual Parallelogram& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,dz),angle));}
    //! apply a homothety on a Parallelogram
    virtual Parallelogram& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.)
      {return transform(Homothety(c,factor));}
    //! apply a homothety on a Parallelogram
    virtual Parallelogram& homothetize(real_t factor)
     {return transform(Homothety(Point(0.,0.,0.),factor));}
    //! apply a point reflection on a Parallelogram
    virtual Parallelogram& pointReflect(const Point& c = Point(0.,0.,0.))
       {return transform(PointReflection(c));}
    //! apply a reflection2d on a Parallelogram
    virtual Parallelogram& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.))
       {return transform(Reflection2d(c,d));}
    //! apply a reflection2d on a Parallelogram
    virtual Parallelogram& reflect2d(const Point& c, real_t dx, real_t dy = 0.)
      {return transform(Reflection2d(c,Point(dx,dy)));}
    //! apply a reflection3d on a Parallelogram
    virtual Parallelogram& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.))
       {return transform(Reflection3d(c,n));}
    //! apply a reflection3d on a Parallelogram
    virtual Parallelogram& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.)
     {return transform(Reflection3d(c,Point(nx,ny,nz)));}

    Vector<real_t> funParametrization(const Point& pt, Parameters& pars, DiffOpType d=_id) const; //!< parametrization (1-u-v)*p1+u*p2+v*p4
    Vector<real_t> invParametrization(const Point& pt, Parameters& pars, DiffOpType d=_id) const; //!< inverse of parametrization (u,v)=invf(p)
};

inline Vector<real_t> parametrization_Parallelogram(const Point& pt, Parameters& pars, DiffOpType d=_id)    //! extern parametrization call
{return reinterpret_cast<const Parallelogram*>(pars("geometry").get_p())->funParametrization(pt,pars,d);}
inline Vector<real_t> invParametrization_Parallelogram(const Point& pt, Parameters& pars, DiffOpType d=_id) //! extern invParametrization call
{return reinterpret_cast<const Parallelogram*>(pars("geometry").get_p())->invParametrization(pt,pars,d);}

//============================================================================================
/*!
  \class Rectangle
  definition of a rectangular geometry in R^3

  Rectangle constructors are based on a key-value system. Here are the available keys:
  - _v1, _v2, _v4: to define the vertices of the Rectangle, counterclockwise oriented
  - _center: to define the center of the Rectangle
  - _origin: to define the first vertex of the Rectangle (same definition as _v1)
  - _xlength, _ylength: the lengths of the Rectangle
  - _xmin, _xmax, _ymin, _ymax: to define the Rectangle v1=(xmin, ymin), v2=(xmax, ymin), v4=(xmin, ymax)
  - _nnodes: to define the number of nodes on each edge of the Rectangle
  - _hsteps: to define the local mesh steps on the vertices of the Rectangle
  - _domain_name: to define the domain name
  - _side_names: to define the side names
  - _varnames: to define the variable names for print purpose
 */
//============================================================================================
class Rectangle : public Parallelogram
{
  protected:
    Point center_;   //!< center of the rectangle
    Point origin_;   //!< "bottom left" vertex of the rectangle (is equal to p_[0])
    bool isCenter_;  //!< true when center key is used
    bool isOrigin_;  //!< true when origin key is used
    real_t xlength_; //!< first length of the rectangle
    real_t ylength_; //!< second length of the rectangle
  private:
    // initialized when rectangle defined with xmin, xmax, ymin, ymax keys
    real_t xmin_, xmax_, ymin_, ymax_;
    // true when key group (_xmin, _xmax, _ymin, _ymax) is given
    bool isBounds_;

  public:
    //! default constructor
    Rectangle();
    //! default constructor with 3 points
    Rectangle(const Point& p1, const Point& p2, const Point& p4,
              const std::vector<number_t>& n=std::vector<number_t>(4,2), const string_t& domName=string_t());
    //! default constructor with 3 points
    Rectangle(const Point& p1, const Point& p2, const Point& p4,
              const std::vector<real_t>& h, const string_t& domName=string_t());
    //! constructor with 3 Parameter
    Rectangle(Parameter p1, Parameter p2, Parameter p3);
    //! constructor with 4 Parameter
    Rectangle(Parameter p1, Parameter p2, Parameter p3, Parameter p4);
    //! constructor with 5 Parameter
    Rectangle(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5);
    //! constructor with 6 Parameter
    Rectangle(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5,
              Parameter p6);
    //! constructor with 7 Parameter
    Rectangle(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5,
              Parameter p6, Parameter p7);

  protected:
    //@{
    //! true constructor functions
    void buildP();
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    //! destructor
    virtual ~Rectangle() {}

    //! returns the number of nodes on first or third edge
    number_t nx() const {return n_[0];}
    //! returns the number of nodes on second or fourth edge
    number_t ny() const {return n_[1];}

    //! format as string
    virtual string_t asString() const;

    virtual Geometry* clone() const { return new Rectangle(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Rectangle(*this); } //!< virtual copy constructor for Surface
    virtual Polygon* clonePG() const { return new Rectangle(*this); } //!< virtual copy constructor for Polygon

    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > surfs() const; //!< returns list of surfaces (const)
    real_t xlength() const { return xlength_; }
    real_t ylength() const { return ylength_; }
    real_t measure() const { return (xlength_ * ylength_); }

    virtual void computeMB() { minimalBox= MinimalBox(p_[0],p_[1],p_[3]); } //!< computes the minimal box

    //! access to child Rectangle object (const)
    virtual const Rectangle* rectangle() const {return this;}
    //! access to child Rectangle object
    virtual Rectangle* rectangle() {return this;}

    //------------------------------------------------
    //         transformations facilities
    //------------------------------------------------
   //! apply a geometrical transformation on a Rectangle
    virtual Rectangle& transform(const Transformation& t);
    //! apply a translation on a Rectangle (vector version)
    virtual Rectangle& translate(std::vector<real_t> u = std::vector<real_t>(3,0.))
      {return transform(Translation(u));}
    //! apply a translation on a Rectangle (3 reals version)
    virtual Rectangle& translate(real_t ux, real_t uy = 0., real_t uz = 0.)
      {return transform(Translation(Point(ux,uy,uz)));}
    //! apply a rotation on a Rectangle
    virtual Rectangle& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.)
      {return transform(Rotation2d(c,angle));}
    //! apply a rotation on a Rectangle
    virtual Rectangle& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.)
      {return transform(Rotation3d(c,d,angle));}
    //! apply a rotation on a Rectangle
    virtual Rectangle& rotate3d(real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,0.),angle));}
    //! apply a rotation on a Rectangle
    virtual Rectangle& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,dz),angle));}
    //! apply a rotation on a Rectangle
    virtual Rectangle& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,0.),angle));}
    //! apply a rotation on a Rectangle
    virtual Rectangle& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,dz),angle));}
    //! apply a homothety on a Rectangle
    virtual Rectangle& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.)
      {return transform(Homothety(c,factor));}
    //! apply a homothety on a Rectangle
    virtual Rectangle& homothetize(real_t factor)
     {return transform(Homothety(Point(0.,0.,0.),factor));}
    //! apply a point reflection on a Rectangle
    virtual Rectangle& pointReflect(const Point& c = Point(0.,0.,0.))
       {return transform(PointReflection(c));}
    //! apply a reflection2d on a Rectangle
    virtual Rectangle& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.))
       {return transform(Reflection2d(c,d));}
    //! apply a reflection2d on a Rectangle
    virtual Rectangle& reflect2d(const Point& c, real_t dx, real_t dy = 0.)
      {return transform(Reflection2d(c,Point(dx,dy)));}
    //! apply a reflection3d on a Rectangle
    virtual Rectangle& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.))
       {return transform(Reflection3d(c,n));}
    //! apply a reflection3d on a Rectangle
    virtual Rectangle& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.)
     {return transform(Reflection3d(c,Point(nx,ny,nz)));}
};

//============================================================================================
/*!
  \class SquareGeo
  definition of a square geometry in R^3

  SquareGeo constructors are based on a key-value system. Here are the available keys:
  - _v1, _v2, _v4: to define the vertices of the Square, counterclockwise oriented
  - _center: to define the center of the SquareGeo
  - _origin: to define the first vertex of the SquareGeo (same definition as _v1)
  - _length: the length of the SquareGeo
  - _xmin, _xmax, _ymin, _ymax: to define the SquareGeo v1=(xmin, ymin), v2=(xmax, ymin), v4=(xmin, ymax)
  - _nnodes: to define the number of nodes on each edge of the SquareGeo
  - _hsteps: to define the local mesh steps on the vertices of the SquareGeo
  - _domain_name: to define the domain name
  - _side_names: to define the side names
  - _varnames: to define the variable names for print purpose
 */
class SquareGeo : public Rectangle
{
  public:
    //! default constructor
    SquareGeo();
    //! default constructor with 3 points
    SquareGeo(const Point& p1, const Point& p2, const Point& p4,
           const std::vector<number_t>& n=std::vector<number_t>(4,2), const string_t& domName=string_t());
    //! default constructor with 3 points
    SquareGeo(const Point& p1, const Point& p2, const Point& p4,
           const std::vector<real_t>& h, const string_t& domName=string_t());
    //! constructor with 2 Parameter
    SquareGeo(Parameter p1, Parameter p2);
    //! constructor with 3 Parameter
    SquareGeo(Parameter p1, Parameter p2, Parameter p3);
    //! constructor with 4 Parameter
    SquareGeo(Parameter p1, Parameter p2, Parameter p3, Parameter p4);
    //! constructor with 5 Parameter
    SquareGeo(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5);
    //! constructor with 6 Parameter
    SquareGeo(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5,
           Parameter p6);

  private:
    //@{
    //! true constructor functions
    void buildP();
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    //! destructor
    virtual ~SquareGeo() {}

    //! format as string
    virtual string_t asString() const;

    virtual Geometry* clone() const { return new SquareGeo(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new SquareGeo(*this); } //!< virtual copy constructor for Surface
    virtual Polygon* clonePG() const { return new SquareGeo(*this); } //!< virtual copy constructor for Polygon

    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > surfs() const; //!< returns list of surfaces (const)

    //! access to child SquareGeo object (const)
    virtual const SquareGeo* square() const {return this;}
    //! access to child SquareGeo object
    virtual SquareGeo* square() {return this;}

    //------------------------------------------------
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a SquareGeo
    virtual SquareGeo& transform(const Transformation& t);
    //! apply a translation on a SquareGeo (vector version)
    virtual SquareGeo& translate(std::vector<real_t> u = std::vector<real_t>(3,0.))
      {return transform(Translation(u));}
    //! apply a translation on a SquareGeo (3 reals version)
    virtual SquareGeo& translate(real_t ux, real_t uy = 0., real_t uz = 0.)
      {return transform(Translation(Point(ux,uy,uz)));}
    //! apply a rotation on a SquareGeo
    virtual SquareGeo& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.)
      {return transform(Rotation2d(c,angle));}
    //! apply a rotation on a SquareGeo
    virtual SquareGeo& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.)
      {return transform(Rotation3d(c,d,angle));}
    //! apply a rotation on a SquareGeo
    virtual SquareGeo& rotate3d(real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,0.),angle));}
    //! apply a rotation on a SquareGeo
    virtual SquareGeo& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,dz),angle));}
    //! apply a rotation on a SquareGeo
    virtual SquareGeo& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,0.),angle));}
    //! apply a rotation on a SquareGeo
    virtual SquareGeo& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,dz),angle));}
    //! apply a homothety on a SquareGeo
    virtual SquareGeo& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.)
      {return transform(Homothety(c,factor));}
    //! apply a homothety on a SquareGeo
    virtual SquareGeo& homothetize(real_t factor)
     {return transform(Homothety(Point(0.,0.,0.),factor));}
    //! apply a point reflection on a SquareGeo
    virtual SquareGeo& pointReflect(const Point& c = Point(0.,0.,0.))
       {return transform(PointReflection(c));}
    //! apply a reflection2d on a SquareGeo
    virtual SquareGeo& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.))
       {return transform(Reflection2d(c,d));}
    //! apply a reflection2d on a SquareGeo
    virtual SquareGeo& reflect2d(const Point& c, real_t dx, real_t dy = 0.)
      {return transform(Reflection2d(c,Point(dx,dy)));}
    //! apply a reflection3d on a SquareGeo
    virtual SquareGeo& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.))
       {return transform(Reflection3d(c,n));}
    //! apply a reflection3d on a SquareGeo
    virtual SquareGeo& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.)
     {return transform(Reflection3d(c,Point(nx,ny,nz)));}
};

/*!
  \class Ellipse
  definition of an elliptic geometry in R^3 (surface)

  Ellipse constructors are based on a key-value system. Here are the available keys:
  - _center: to define the center of the Ellipse
  - _v1, _v2: to define apogees of the Ellipse
  - _xlength, _ylength: to define axis lengths of the Ellipse
  - _xradius, _yradius: to define semi-axis lengths of the Ellipse
  - _angle1, _angle2: to define an elliptic sector from a pair of angles. The angular origin is determined by _v1
  - _type: indicator to fit curved boundaries (default) or not which gives flat (or plane) boundaries
  - _nnodes: to define the number of nodes on the edges of the Ellipse
  - _hsteps: to define the local mesh steps on build points of the Ellipse
  - _domain_name: to define the domain name
  - _side_names: to define the side names
  - _varnames: to define the variable names for print purpose
 */
//============================================================================================
class Ellipse : public Surface
{
  protected:
    /*!
      vertices of the ellipse
      - p_[0] is the center of the ellipse
      - p_[1] first apogee of the ellipse
      - p_[2] second apogee of the ellipse
      - p_[3] first perigee of the ellipse
      - p_[4] second perigee of the ellipse
    */

    std::vector<Point> v_;    //!< true list of vertices according to number of octants
    real_t xradius_, yradius_;
    bool isAxis_;
    real_t thetamin_;       //!< first angle defining the sector, given in radian
    real_t thetamax_;       //!< second angle defining the sector, given in radian
    bool isSector_;         //!< true if thetamin_ > 0 or thetamax_ < 2pi
    dimen_t type_;          //!< indicator to fit curved boundaries (default) or not which gives flat (or plane) boundaries

  public:
    //! default constructor
    Ellipse();
    //! default constructor with 3 points
    Ellipse(const Point& center, const Point& p1, const Point& p2, const std::vector<number_t>& n = std::vector<number_t>(4,2),
            const string_t& domName=string_t());
    //! default constructor with 3 points
    Ellipse(const Point& center, const Point& p1, const Point& p2, const std::vector<real_t>& h, const string_t& domName=string_t());
    //! constructor with 3 Parameter
    Ellipse(Parameter p1, Parameter p2, Parameter p3);
    //! constructor with 4 Parameter
    Ellipse(Parameter p1, Parameter p2, Parameter p3, Parameter p4);
    //! constructor with 5 Parameter
    Ellipse(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5);
    //! constructor with 6 Parameter
    Ellipse(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5,
            Parameter p6);
    //! constructor with 7 Parameter
    Ellipse(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5,
            Parameter p6, Parameter p7);
    //! constructor with 8 Parameter
    Ellipse(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5,
            Parameter p6, Parameter p7, Parameter p8);
    //! constructor with 9 Parameter
    Ellipse(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5,
            Parameter p6, Parameter p7, Parameter p8, Parameter p9);

  protected:
    //@{
    //! true constructor functions
    void buildPVNAndH();
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    virtual void setParametrization();
    //@}

  public:
    //! destructor
    virtual ~Ellipse() {}

    //! accessor to point
    Point center() const { return p_[0]; }
    real_t length1() const { return 2.*xradius_; } //!< return first axis length
    real_t length2() const { return 2.*yradius_; } //!< return second axis length
    real_t radius1() const { return xradius_; }    //!< return first semi-axis length
    real_t radius2() const { return yradius_; }    //!< return second semi-axis length
    virtual number_t nbSubdiv() const;             //!< returns number of subdivision
    virtual dimen_t type() const { return type_; } //!< returns type of subdivision
    real_t thetamin() const { return thetamin_; }  //!< returns the first angle
    real_t thetamax() const { return thetamax_; }  //!< returns the second angle
    bool isSector() const { return isSector_; }    //!< returns the flag being true if it is an elliptical sector
    virtual bool isPlane() const {return true;}

    //! accessor to true vertices
    const std::vector<Point>& v() const { return v_; }
    //! accessor to vertex i (read only)
    const Point& v(number_t i) const { return v_[i-1]; }

    //! format as string
    virtual string_t asString() const;

    virtual Geometry* clone() const { return new Ellipse(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Ellipse(*this); } //!< virtual copy constructor for Surface

    std::vector<int_t> nnodesPerBorder(); //!< i-th value is nnodes of the Geometry's i-th border, else -1 if defined with hsteps

    virtual std::vector<const Point*> boundNodes() const; //!< returns list of points on boundary (const)
    virtual std::vector<Point*> nodes(); //!< returns list of every main point (non const)
    virtual std::vector<Point*> wholeNodes(); //!< returns list of every point (p_ and v_) (non const)
    virtual std::vector<const Point*> nodes() const; //!< returns list of every point (const)
    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > curves() const; //!< returns list of edges (const)
    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > surfs() const; //!< returns list of surfaces (const)
    virtual number_t nbSides() const { return 4; } //!< returns the number of sides
    real_t measure() const { return pi_*xradius_*yradius_; }

    virtual void collect(const string_t& n, std::list<Geometry*>&) const; //!< collect in a list all canonical geometry's with name n

    //! computes the bounding box
    virtual void computeBB();
    //! computes the minimal box
    virtual void computeMB();

    //! access to child Ellipse object (const)
    virtual const Ellipse* ellipse() const {return this;}
    //! access to child Ellipse object
    virtual Ellipse* ellipse() {return this;}
     //! access to child Disk object (const)
    virtual const Disk* disk() const
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _disk)); return 0; }
    //! access to child Disk object
    virtual Disk* disk()
    { error("bad_geometry", asString(), words("shape", shape_), words("shape", _disk)); return 0; }
    //! create boundary geometry
    virtual Geometry& buildBoundary() const;

    //------------------------------------------------
    //         transformations facilities
    //------------------------------------------------
    //! apply a geometrical transformation on a Ellipse
    virtual Ellipse& transform(const Transformation& t);
    //! apply a translation on a Ellipse (vector version)
    virtual Ellipse& translate(std::vector<real_t> u = std::vector<real_t>(3,0.))
      {return transform(Translation(u));}
    //! apply a translation on a Ellipse (3 reals version)
    virtual Ellipse& translate(real_t ux, real_t uy = 0., real_t uz = 0.)
      {return transform(Translation(Point(ux,uy,uz)));}
    //! apply a rotation on a Ellipse
    virtual Ellipse& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.)
      {return transform(Rotation2d(c,angle));}
    //! apply a rotation on a Ellipse
    virtual Ellipse& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.)
      {return transform(Rotation3d(c,d,angle));}
    //! apply a rotation on a Ellipse
    virtual Ellipse& rotate3d(real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,0.),angle));}
    //! apply a rotation on a Ellipse
    virtual Ellipse& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,dz),angle));}
    //! apply a rotation on a Ellipse
    virtual Ellipse& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,0.),angle));}
    //! apply a rotation on a Ellipse
    virtual Ellipse& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,dz),angle));}
    //! apply a homothety on a Ellipse
    virtual Ellipse& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.)
      {return transform(Homothety(c,factor));}
    //! apply a homothety on a Ellipse
    virtual Ellipse& homothetize(real_t factor)
     {return transform(Homothety(Point(0.,0.,0.),factor));}
    //! apply a point reflection on a Ellipse
    virtual Ellipse& pointReflect(const Point& c = Point(0.,0.,0.))
       {return transform(PointReflection(c));}
    //! apply a reflection2d on a Ellipse
    virtual Ellipse& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.))
       {return transform(Reflection2d(c,d));}
    //! apply a reflection2d on a Ellipse
    virtual Ellipse& reflect2d(const Point& c, real_t dx, real_t dy = 0.)
      {return transform(Reflection2d(c,Point(dx,dy)));}
    //! apply a reflection3d on a Ellipse
    virtual Ellipse& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.))
       {return transform(Reflection3d(c,n));}
    //! apply a reflection3d on a Ellipse
    virtual Ellipse& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.)
     {return transform(Reflection3d(c,Point(nx,ny,nz)));}

    Vector<real_t> funParametrization(const Point& pt, Parameters& pars, DiffOpType d=_id) const; //!< parametrization P0+(P1-P0)*r*cos(t)+P2-P0)*r*sin(t)
    Vector<real_t> invParametrization(const Point& pt, Parameters& pars, DiffOpType d=_id) const; //!< inverse of parametrization (u,v)=invf(p)
};

inline Vector<real_t> parametrization_Ellipse(const Point& pt, Parameters& pars, DiffOpType d=_id)    //! extern parametrization call
{return reinterpret_cast<const Ellipse*>(pars("geometry").get_p())->funParametrization(pt,pars,d);}
inline Vector<real_t> invParametrization_Ellipse(const Point& pt, Parameters& pars, DiffOpType d=_id) //! extern invParametrization call
{return reinterpret_cast<const Ellipse*>(pars("geometry").get_p())->invParametrization(pt,pars,d);}

//============================================================================================
/*!
  \class Disk
  definition of a circular geometry in R^3 (surface)

  Disk constructors are based on a key-value system. Here are the available keys:
  - _center: to define the center of the Disk
  - _v1, _v2: to define "apogees" of the Disk
  - _radius: to define semi-axis lengths of the Disk
  - _angle1, _angle2: to define an circular sector from a pair of angles. The angular origin is determined by _v1
  - _type: indicator to fit curved boundaries (default) or not which gives flat (or plane) boundaries
  - _nnodes: to define the number of nodes on the edges of the Disk
  - _hsteps: to define the local mesh steps on build points of the Disk
  - _domain_name: to define the domain name
  - _side_names: to define the side names
  - _varnames: to define the variable names for print purpose
 */
//============================================================================================
class Disk : public Ellipse
{
  public:
    //! default constructor
    Disk();
    //! default constructor with 3 points
    Disk(const Point& center, const Point& p1, const Point& p2, const std::vector<number_t>& n=std::vector<number_t>(4,2),
         const string_t& domName=string_t());
    //! default constructor with 3 points
    Disk(const Point& center, const Point& p1, const Point& p2, const std::vector<real_t>& h, const string_t& domName=string_t());
    //! constructor with 2 Parameter
    Disk(Parameter p1, Parameter p2);
    //! constructor with 3 Parameter
    Disk(Parameter p1, Parameter p2, Parameter p3);
    //! constructor with 4 Parameter
    Disk(Parameter p1, Parameter p2, Parameter p3, Parameter p4);
    //! constructor with 5 Parameter
    Disk(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5);
    //! constructor with 6 Parameter
    Disk(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5,
            Parameter p6);
    //! constructor with 7 Parameter
    Disk(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5,
            Parameter p6, Parameter p7);
    //! constructor with 8 Parameter
    Disk(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5,
            Parameter p6, Parameter p7, Parameter p8);
    //! constructor with 9 Parameter
    Disk(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5,
            Parameter p6, Parameter p7, Parameter p8, Parameter p9);

  protected:
    //@{
    //! true constructor functions
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    //! destructor
    virtual ~Disk() {}

    //! format as string
    virtual string_t asString() const;

    //! computes the bounding box
    // virtual void computeBB();
    //! computes the minimal box
    // virtual void computeMB();

    virtual Geometry* clone() const { return new Disk(*this); } //!< virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new Disk(*this); } //!< virtual copy constructor for Surface

    //virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > curves() const; //!< returns list of edges

    //! access to child Disk object (const)
    virtual const Disk* disk() const {return this;}
    //! access to child Disk object
    virtual Disk* disk() {return this;}
    real_t radius() const { return xradius_; }    //!< return radius

    //------------------------------------------------
    //         transformations facilities
    //------------------------------------------------
    //! apply a geometrical transformation on a Disk
    virtual Disk& transform(const Transformation& t);
    //! apply a translation on a Disk (vector version)
    virtual Disk& translate(std::vector<real_t> u = std::vector<real_t>(3,0.))
      {return transform(Translation(u));}
    //! apply a translation on a Disk (3 reals version)
    virtual Disk& translate(real_t ux, real_t uy = 0., real_t uz = 0.)
      {return transform(Translation(Point(ux,uy,uz)));}
    //! apply a rotation on a Disk
    virtual Disk& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.)
      {return transform(Rotation2d(c,angle));}
    //! apply a rotation on a Disk
    virtual Disk& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.)
      {return transform(Rotation3d(c,d,angle));}
    //! apply a rotation on a Disk
    virtual Disk& rotate3d(real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,0.),angle));}
    //! apply a rotation on a Disk
    virtual Disk& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,dz),angle));}
    //! apply a rotation on a Disk
    virtual Disk& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,0.),angle));}
    //! apply a rotation on a Disk
    virtual Disk& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,dz),angle));}
    //! apply a homothety on a Disk
    virtual Disk& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.)
      {return transform(Homothety(c,factor));}
    //! apply a homothety on a Disk
    virtual Disk& homothetize(real_t factor)
     {return transform(Homothety(Point(0.,0.,0.),factor));}
    //! apply a point reflection on a Disk
    virtual Disk& pointReflect(const Point& c = Point(0.,0.,0.))
       {return transform(PointReflection(c));}
    //! apply a reflection2d on a Disk
    virtual Disk& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.))
       {return transform(Reflection2d(c,d));}
    //! apply a reflection2d on a Disk
    virtual Disk& reflect2d(const Point& c, real_t dx, real_t dy = 0.)
      {return transform(Reflection2d(c,Point(dx,dy)));}
    //! apply a reflection3d on a Disk
    virtual Disk& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.))
       {return transform(Reflection3d(c,n));}
    //! apply a reflection3d on a Disk
    virtual Disk& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.)
     {return transform(Reflection3d(c,Point(nx,ny,nz)));}
};

typedef Disk Circle;

//===================================================================================================
/*!
  \class ParametrizedSurface
  definition of a ParametrizedSurface geometry in R^3 : x1(u,v), x2(u,v), x3(u,v)
  As gmsh does not support parametrized surface, to be exported, ParametrizedSurface can be split into _nbparts parts
    - smaller quadrangles or triangles (_linearPartition)
    - smaller nurbs (_splinePartition)  NOT YET AVAILABLE

  ParametrizedSurface constructors are based on a key-value system. Here are the available keys:
  - _parametrization: parametrization object is handled by Geometry
  - _partitioning: type of partition, one of _nonePartition*, _linearPartition, _splinePartition
  - _nbparts: number of partitions
  - _nnodes: to define the number of nodes on the bound edges
  - _hsteps: to define the local mesh steps on the bounds nodes
  - _domain_name: to define the domain name
  - _side_names: to define the side names
  - _varnames: to define the variable names for print purpose
*/
//===================================================================================================
class ParametrizedSurface : public Surface
{
  protected:
    Partitioning partitioning_; //!< partitioning type  (one of _nonePartition*, _linearPartition _splinePartition)
    number_t nbParts_;          //!< number of partitions (>0)
    Transformation* transformation_;  //!< pointer to an additional linear transformation to apply to (0 by default: no additional linear transformation)

  public:
    //! default constructor
    ParametrizedSurface();
    //! constructor with 2 Parameter
    ParametrizedSurface(Parameter p1, Parameter p2);
    //! constructor with 3 Parameter
    ParametrizedSurface(Parameter p1, Parameter p2, Parameter p3);
    //! constructor with 4 Parameter
    ParametrizedSurface(Parameter p1, Parameter p2, Parameter p3, Parameter p4);
    //! constructor with 5 Parameter
    ParametrizedSurface(Parameter p1, Parameter p2, Parameter p3, Parameter p4,
            Parameter p5);
    ParametrizedSurface(Parameter p1, Parameter p2, Parameter p3, Parameter p4,
            Parameter p5, Parameter p6);

  private:
    //@{
    //! true constructor functions
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    ParametrizedSurface(const ParametrizedSurface&);                             //!< copy constructor
    virtual Geometry* clone() const { return new ParametrizedSurface(*this); }  //! virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new ParametrizedSurface(*this); }  //!< virtual copy constructor for Surface
    //! destructor
    virtual ~ParametrizedSurface() {if(transformation_!=0) delete transformation_;}

    //! return Partitioning
    Partitioning partitioning() const {return partitioning_;}
    //! return nb of partitions
    number_t nbParts() const {return nbParts_;}
    //! return pointer to mesh related to geometry support of parametrization
    const Mesh& supportMesh() const {return *parametrization_->meshP();}

    //! format as string
    virtual string_t asString() const;

     //! access to child ParametrizedSurface object (const)
    virtual const ParametrizedSurface* parametrizedSurface() const {return this;}
    //! access to child ParametrizedSurface object
    virtual ParametrizedSurface* parametrizedSurface() {return this;}

    virtual number_t nbSides() const { return 4; }                //!< returns the number of sides
    virtual std::vector<const Point*> nodes() const;              //!< returns the nodes (including internal nodes if partition)
    virtual std::vector<const Point*> boundNodes() const;         //!< returns the nodes on boundary (including internal nodes if partition)
    std::vector<const Point*> nodesOnSide(number_t s=0) const;    //!< return the nodes located on the side s>=1 (any side if s=0)
    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > curves() const;
    virtual std::vector<std::pair<ShapeType,std::vector<const Point*> > > surfs() const;
    virtual Geometry& buildBoundary() const;                      //!< create boundary geometry of ParametrizedSurface (composite of ArcParametrization)

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a ParametrizedSurface
    virtual ParametrizedSurface& transform(const Transformation& t);
    //! apply a translation on a ParametrizedSurface (vector version)
    virtual ParametrizedSurface& translate(std::vector<real_t> u = std::vector<real_t>(3,0.))
      {return transform(Translation(u));}
    //! apply a translation on a ParametrizedSurface (3 reals version)
    virtual ParametrizedSurface& translate(real_t ux, real_t uy = 0., real_t uz = 0.)
      {return transform(Translation(Point(ux,uy,uz)));}
    //! apply a rotation on a ParametrizedSurface
    virtual ParametrizedSurface& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.)
      {return transform(Rotation2d(c,angle));}
    //! apply a rotation on a ParametrizedSurface
    virtual ParametrizedSurface& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.)
      {return transform(Rotation3d(c,d,angle));}
    //! apply a rotation on a ParametrizedSurface
    virtual ParametrizedSurface& rotate3d(real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,0.),angle));}
    //! apply a rotation on a ParametrizedSurface
    virtual ParametrizedSurface& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,dz),angle));}
    //! apply a rotation on a ParametrizedSurface
    virtual ParametrizedSurface& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,0.),angle));}
    //! apply a rotation on a ParametrizedSurface
    virtual ParametrizedSurface& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,dz),angle));}
    //! apply a homothety on a ParametrizedSurface
    virtual ParametrizedSurface& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.)
      {return transform(Homothety(c,factor));}
    //! apply a homothety on a ParametrizedSurface
    virtual ParametrizedSurface& homothetize(real_t factor)
     {return transform(Homothety(Point(0.,0.,0.),factor));}
    //! apply a point reflection on a ParametrizedSurface
    virtual ParametrizedSurface& pointReflect(const Point& c = Point(0.,0.,0.))
       {return transform(PointReflection(c));}
    //! apply a reflection2d on a ParametrizedSurface
    virtual ParametrizedSurface& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.))
       {return transform(Reflection2d(c,d));}
    //! apply a reflection2d on a ParametrizedSurface
    virtual ParametrizedSurface& reflect2d(const Point& c, real_t dx, real_t dy = 0.)
      {return transform(Reflection2d(c,Point(dx,dy)));}
    //! apply a reflection3d on a ParametrizedSurface
    virtual ParametrizedSurface& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.))
       {return transform(Reflection3d(c,n));}
    //! apply a reflection3d on a ParametrizedSurface
    virtual ParametrizedSurface& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.)
     {return transform(Reflection3d(c,Point(nx,ny,nz)));}

};

//===================================================================================================
/*!
  \class SplineSurface
  definition of a SplineSurface geometry in R^3 from a Nurbs object (non uniform rational Bspline)
  only nurbs are available either approximation nurbs or interpolation nurbs (default when subtype is not specified)

  SplineSurface constructors are based on a key-value system. Here are the available keys:
  - _vertices: list of control points
  - _spline: spline object
  - _spline_type: type of spline one of _C2Spline, _CatmullRomSpline, _BSpline, _BezierSpline, _Nurbs
  - _spline_subtype: subtype of spline, one of _SplineInterpolation, _SplineApproximation
  - _spline_BC: spline boundary condition, one of _naturalBC, _clampedBC, _periodicBC
  - _degree: degree of spline (default 3)
  - _nbu: number of control/interpolation points in u-direction (_Nurbs only)
  - _spline_parametrization: spline parametrization type, one of _xParametrization, _uniformParametrization,_chordalParametrization, _centripetalParametrization
  - _weights: weights of control points (for _BSpline only)
  - _nnodes: to define the number of nodes on the arc
  - _hsteps: to define the local mesh steps on the bounds
  - _domain_name: to define the domain name
  - _side_names: to define the side names
  - _varnames: to define the variable names for print purpose
*/
//===================================================================================================
class SplineSurface : public Surface
{
  protected:
    Spline* spline_;     //!< spline pointer (nurbs)

  public:
    //! default constructor
    SplineSurface();
    //@{
    //! constructor with Parameters
    SplineSurface(Parameter p1);
    SplineSurface(Parameter p1, Parameter p2);
    SplineSurface(Parameter p1, Parameter p2, Parameter p3);
    SplineSurface(Parameter p1, Parameter p2, Parameter p3, Parameter p4);
    SplineSurface(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5);
    SplineSurface(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5, Parameter p6);
    SplineSurface(Parameter p1, Parameter p2, Parameter p3, Parameter p4, Parameter p5, Parameter p6, Parameter p7);
    //@}

  private:
    //@{
    //! true constructor functions
    void build(const std::vector<Parameter>& ps);
    void buildParam(const Parameter& gp);
    void buildDefaultParam(ParameterKey key);
    std::set<ParameterKey> getParamsKeys();
    //@}

  public:
    SplineSurface(const SplineSurface&);                                  //!< copy constructor
    void copy(const SplineSurface&);                                      //!< real copy
    SplineSurface& operator=(const SplineSurface&);                       //!< assign operator
    virtual Geometry* clone() const { return new SplineSurface(*this); }  //! virtual copy constructor for Geometry
    virtual Surface* cloneS() const { return new SplineSurface(*this); }  //!< virtual copy constructor for Surface
    virtual ~SplineSurface();

    const Spline* spline() const {return spline_;};
     //! access to child SplineSurface object (const)
    virtual const SplineSurface* splineSurface() const {return this;}
    //! access to child SplineSurface object
    virtual SplineSurface* splineSurface() {return this;}
    string_t asString() const; //!< format as string
    virtual Geometry& buildBoundary() const;      //!< create boundary geometry of SplineSurface (composite of SplineArc (BSpline))

    //================================================
    //         transformations facilities
    //================================================
    //! apply a geometrical transformation on a SplineSurface
    virtual SplineSurface& transform(const Transformation& t);
    //! apply a translation on a SplineSurface (vector version)
    virtual SplineSurface& translate(std::vector<real_t> u = std::vector<real_t>(3,0.))
      {return transform(Translation(u));}
    //! apply a translation on a SplineSurface (3 reals version)
    virtual SplineSurface& translate(real_t ux, real_t uy = 0., real_t uz = 0.)
      {return transform(Translation(Point(ux,uy,uz)));}
    //! apply a rotation on a SplineSurface
    virtual SplineSurface& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.)
      {return transform(Rotation2d(c,angle));}
    //! apply a rotation on a SplineSurface
    virtual SplineSurface& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.)
      {return transform(Rotation3d(c,d,angle));}
    //! apply a rotation on a SplineSurface
    virtual SplineSurface& rotate3d(real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,0.),angle));}
    //! apply a rotation on a SplineSurface
    virtual SplineSurface& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,dz),angle));}
    //! apply a rotation on a SplineSurface
    virtual SplineSurface& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,0.),angle));}
    //! apply a rotation on a SplineSurface
    virtual SplineSurface& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,dz),angle));}
    //! apply a homothety on a SplineSurface
    virtual SplineSurface& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.)
      {return transform(Homothety(c,factor));}
    //! apply a homothety on a SplineSurface
    virtual SplineSurface& homothetize(real_t factor)
     {return transform(Homothety(Point(0.,0.,0.),factor));}
    //! apply a point reflection on a SplineSurface
    virtual SplineSurface& pointReflect(const Point& c = Point(0.,0.,0.))
       {return transform(PointReflection(c));}
    //! apply a reflection2d on a SplineSurface
    virtual SplineSurface& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.))
       {return transform(Reflection2d(c,d));}
    //! apply a reflection2d on a SplineSurface
    virtual SplineSurface& reflect2d(const Point& c, real_t dx, real_t dy = 0.)
      {return transform(Reflection2d(c,Point(dx,dy)));}
    //! apply a reflection3d on a SplineSurface
    virtual SplineSurface& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.))
       {return transform(Reflection3d(c,n));}
    //! apply a reflection3d on a SplineSurface
    virtual SplineSurface& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.)
     {return transform(Reflection3d(c,Point(nx,ny,nz)));}

    Vector<real_t> funParametrization(const Point& pt, Parameters& pars, DiffOpType d=_id) const; //!< interface to spline parametrization
    Vector<real_t> invParametrization(const Point& pt, Parameters& pars, DiffOpType d=_id) const; //!< interface to spline parametrization
};

inline Vector<real_t> parametrization_SplineSurface(const Point& pt, Parameters& pars, DiffOpType d=_id)    //! extern parametrization call
{return reinterpret_cast<const SplineSurface*>(pars("geometry").get_p())->funParametrization(pt,pars,d);}
inline Vector<real_t> invParametrization_SplineSurface(const Point& pt, Parameters& pars, DiffOpType d=_id) //! extern invParametrization call
{return reinterpret_cast<const SplineSurface*>(pars("geometry").get_p())->invParametrization(pt,pars,d);}


//============================================================================================
/*!
   \class SetOfElems
   This class defines an initial mesh from a list of points (pts), a list of elements (elems)
   and a list of boundaries (bounds).
     Each element is defined by N numbers corresponding to the rank of its N vertices
   in the vector pts ; the numbers start from 1. Thus, elems[k][i] is the number of the
   vertex (i+1) of the element number (k+1) (if the elements are numbered starting
   from 1), and 1 <= elems[k][i] <= pts.size().
     The boundaries of the polygonal domain defined by the elements are also defined
   by point numbers with the same convention. Each boundary is defined by the list of
   numbers of the points they belong to, given in any order.
     The last argument nbsubdiv is the number of subdivisions wanted. The initial
   mesh corresponds to nbsubdiv=0.
   -> Important: the subdivision algorithm generate a new point on each edge of the mesh.
      It belongs to boundary B if the two end points of the edge belong to the boundary B ;
      it is an internal point otherwise. Owing to this rule, in order to get a correct result,
      care must be taken in the definition of the boundaries since too few boundaries may
      lead to get boundary points that are in fact inside the domain. For instance,
      subdividing one single triangle leads to mandatory define its three edges as boundaries.
*/
//============================================================================================
class SetOfElems : public Geometry
{
  private:
    std::vector<Point> pts_; //!< list of points
    const std::vector<std::vector<number_t> >& elems_; //!< list of elements (an element is defined by its list of nodes)
    const std::vector<std::vector<number_t> >& bounds_; //!< list of elements on boundary
    ShapeType elemShape_;             //!< shape of the elements in the mesh
    number_t nbsubdiv_;           //!< number of subdivisions to be performed to get the final mesh
    const number_t nbElems_, nbBounds_;

  public:
    //! default constructor
    SetOfElems(const std::vector<Point>& pts, const std::vector<std::vector<number_t> >& elems,
               const std::vector<std::vector<number_t> >& bounds, const std::vector<string_t>& names,
               const ShapeType esh, const number_t nbsubdiv=1)
    : Geometry(BoundingBox(pts), BoundingBox(pts).dim(), string_t("Omega"), _setofelems),
      pts_(pts), elems_(elems), bounds_(bounds), elemShape_(esh), nbsubdiv_(nbsubdiv),
      nbElems_(elems_.size()), nbBounds_(bounds_.size())
    { sideNames_ = names; }

    //! format as string
    virtual string_t asString() const;

    virtual Geometry* clone() const { return new SetOfElems(*this); } //!< virtual copy constructor

    virtual std::vector<Point*> nodes(); //!< list of every point (non const)
    virtual std::vector<const Point*> nodes() const; //!< list of every point (const)

    // access functions to data members
    std::vector<Point>& pts() { return pts_; } //!< return list of points (non const)
    const std::vector<Point>& pts() const { return pts_; } //!< returns list of points (const)
    const std::vector<std::vector<number_t> >& elems() const { return elems_; } //!< returns list of elements
    const std::vector<std::vector<number_t> >& bounds() const { return bounds_; } //!< returns list of elements on boundary
    ShapeType elemShape() const { return elemShape_; } //!< returns shape of the elements
    virtual number_t nbSubdiv() const { return nbsubdiv_; } //!< returns number of subdivision
    virtual number_t nbSides() const { return bounds_.size(); } //!< returns the number of sides

    //! access to child SetOfElems object (const)
    virtual const SetOfElems* setofelems() const {return this;}
    //! access to child SetOfElems object
    virtual SetOfElems* setofelems() {return this;}

    //------------------------------------------------
    //         transformations facilities
    //------------------------------------------------
    //! apply a geometrical transformation on a SetOfElems
    virtual SetOfElems& transform(const Transformation& t);
    //! apply a translation on a SetOfElems (vector version)
    virtual SetOfElems& translate(std::vector<real_t> u = std::vector<real_t>(3,0.))
      {return transform(Translation(u));}
    //! apply a translation on a SetOfElems (3 reals version)
    virtual SetOfElems& translate(real_t ux, real_t uy = 0., real_t uz = 0.)
      {return transform(Translation(Point(ux,uy,uz)));}
    //! apply a rotation on a SetOfElems
    virtual SetOfElems& rotate2d(const Point& c = Point(0.,0.), real_t angle = 0.)
      {return transform(Rotation2d(c,angle));}
    //! apply a rotation on a SetOfElems
    virtual SetOfElems& rotate3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> d = std::vector<real_t>(3,0.),
                              real_t angle = 0.)
      {return transform(Rotation3d(c,d,angle));}
    //! apply a rotation on a SetOfElems
    virtual SetOfElems& rotate3d(real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,0.),angle));}
    //! apply a rotation on a SetOfElems
    virtual SetOfElems& rotate3d(real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(Point(0.,0.,0.),Point(dx,dy,dz),angle));}
    //! apply a rotation on a SetOfElems
    virtual SetOfElems& rotate3d(const Point& c, real_t dx, real_t dy, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,0.),angle));}
    //! apply a rotation on a SetOfElems
    virtual SetOfElems& rotate3d(const Point& c, real_t dx, real_t dy, real_t dz, real_t angle)
      {return transform(Rotation3d(c,Point(dx,dy,dz),angle));}
    //! apply a homothety on a SetOfElems
    virtual SetOfElems& homothetize(const Point& c = Point(0.,0.,0.), real_t factor = 1.)
      {return transform(Homothety(c,factor));}
    //! apply a homothety on a SetOfElems
    virtual SetOfElems& homothetize(real_t factor)
     {return transform(Homothety(Point(0.,0.,0.),factor));}
    //! apply a point reflection on a SetOfElems
    virtual SetOfElems& pointReflect(const Point& c = Point(0.,0.,0.))
       {return transform(PointReflection(c));}
    //! apply a reflection2d on a SetOfElems
    virtual SetOfElems& reflect2d(const Point& c = Point(0.,0.), std::vector<real_t> d = std::vector<real_t>(2,0.))
       {return transform(Reflection2d(c,d));}
    //! apply a reflection2d on a SetOfElems
    virtual SetOfElems& reflect2d(const Point& c, real_t dx, real_t dy = 0.)
      {return transform(Reflection2d(c,Point(dx,dy)));}
    //! apply a reflection3d on a SetOfElems
    virtual SetOfElems& reflect3d(const Point& c = Point(0.,0.,0.), std::vector<real_t> n = std::vector<real_t>(3,0.))
       {return transform(Reflection3d(c,n));}
    //! apply a reflection3d on a SetOfElems
    virtual SetOfElems& reflect3d(const Point& c, real_t nx, real_t ny, real_t nz = 0.)
     {return transform(Reflection3d(c,Point(nx,ny,nz)));}
};

} // end of namespace xlifepp

#endif // GEOMETRIES_2D_HPP
